# -*- coding: utf-8 -*-
import logging
import re
import time
import urllib2

from tempfile import NamedTemporaryFile
from lxml import etree
from os.path import isfile
from functools import wraps

_logger = logging.getLogger(__name__)


def retry(exception_to_check, tries=4, delay=3, back_off=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param exception_to_check: the exception to check. may be a tuple of
        exceptions to check
    :type exception_to_check: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param back_off: back_off multiplier e.g. value of 2 will double the
        delay each retry
    :type back_off: int
    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance
    """

    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exception_to_check, e:
                    msg = "%s, Retrying in %d seconds..." % (str(e),
                                                             mdelay)
                    if logger:
                        logger.warning(msg)
                    else:
                        print msg
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= back_off
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


class Tools(object):

    cached = {}

    @staticmethod
    def is_url(element_name):
        return element_name.startswith('http')

    @retry(urllib2.URLError, tries=3, delay=3, back_off=2)
    def cache_it(self, url):
        """Take an url which deliver a plain document  and convert it to a
        temporary file, this document is an xslt file expecting contains all
        xslt definitions, then the cache process is recursive.

        :param url: document origin url
        :return file path: local new absolute path
        """
        def cache_it(_url):
            cached = NamedTemporaryFile(delete=False)
            named = cached.name
            _response = urllib2.urlopen(_url)
            _content = _response.read()
            with cached as cache:
                cache.write(_content)
            self.cached.update({_url: named})
            return _content, named

        # If on this runtime it is cached do not cache it again.
        if self.cached.get(url) and isfile(self.cached[url]):
            return self.cached[url]

        internal = 0

        # TODO: unwire this to xslt, but usefull for know
        if url.endswith('xslt') and not internal:

            # Opening the first file just the first time.

            response = urllib2.urlopen(url)
            content = response.read()

            # Find all urls in the main xslt file.

            urls = re.findall(r'href=[\'"]?([^\'" >]+)', content)

            # Mapping all internal url in the file to local cached files.

            for u in urls:
                internal += 1
                content, _file = cache_it(u)
                content.replace(u, _file)
        return cache_it(url)[1]

    @staticmethod
    def get_original(document, xslt):
        """Get the original chain given document path and xslt local path

        :param str document: local absolute path to document
        :param str xslt: local absolute path to xst file
        :return: new chain generated.
        :rtype: str
        """
        dom = etree.parse(document)  # TODO: cuando este probando -
        # fuente:
        # http://stackoverflow.com/questions/16698935/how-to-transform-an-xml-file-using-xslt-in-python
        xslt = etree.parse(xslt)
        transform = etree.XSLT(xslt)
        newdom = transform(dom)
        return newdom


tools = Tools()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_cfdilib
----------------------------------

Tests for `cfdilib` module.
Tests for `cfdv32` module.
"""
import json

from os.path import join, dirname
import unittest

from cfdilib import cfdilib, cfdv32
from cfdilib.tools import tools


class TestCfdilib(unittest.TestCase):

    def _get_test_file(self, file_name):
        test_data_file = join(self.test_data_dir, file_name)
        with open(test_data_file) as data_file:
            return data_file.read()

    def setUp(self):
        self.test_data_dir = join(
            dirname(cfdilib.__file__), "..", "tests", "demo")
        self.dict_invoice_basic_32 = eval(
            self._get_test_file('basic_invoice_32.txt'))
        self.dict_invoice_basic_32_errored = eval(
            self._get_test_file('basic_invoice_32_errored.txt'))
        self.dict_invoice_basic_32_false = eval(
            self._get_test_file('basic_invoice_32_false.txt'))
        self.real_document_xml = join(
            dirname(cfdilib.__file__), "..", "tests", "demo", 'cfdv32.xml')

    def tearDown(self):
        pass

    def test_001_get_xsd_documentation(self):
        """Getting a documentation from a given Clark's Notated xsd element"""
        invoice = cfdv32.get_invoice(self.dict_invoice_basic_32)
        documentation = invoice.get_documentation(
                '{http://www.sat.gob.mx/cfd/3}'
                'Impuestos')
        self.assertTrue(True)  # TODO: Find the proper search element syntax


    def test_002_get_cfd_debugged(self):
        """With a given valid dict an
        invoice object is created in debug_mode"""
        invoice = cfdv32.get_invoice(self.dict_invoice_basic_32,
                                     debug_mode=True)
        self.assertTrue(invoice.document,
                        'A valid dictionary gave error debugged_mode enabled')

    def test_003_get_cfd(self):
        """With a given valid dict an invoice object is created"""
        invoice = cfdv32.get_invoice(self.dict_invoice_basic_32)
        self.assertTrue(invoice.document,
                        'A valid dictionary gave error an error')

    def test_004_get_errored(self):
        """With a given invalid dict raise properly errors on ups object"""
        invoice = cfdv32.get_invoice(self.dict_invoice_basic_32_errored)
        self.assertTrue(bool(invoice.ups),
                        'An invalid dictionary gave a '
                        'valid output, that is wrong.')
        # Ok it failed!, then we assert if
        # the message is the one I expected for.
        self.assertTrue(invoice.ups.message.find('Emisor') > 0,
                        'The expected failed entry Emisor was erroneous.')

        invoice = cfdv32.get_invoice({})
        self.assertTrue(bool(invoice.ups),
                        'An empty dict should give me the validation')

        invoice = cfdv32.get_invoice(self.dict_invoice_basic_32_false)
        self.assertNotIn('False', invoice.ups.message,
                         'Passing a False value return a False string which is'
                         'incorrect  %s ' % invoice.ups.message)

    def test_005_get_cfd_invalid_debugged(self):
        """With a given `invalid` dict an invoice
        object is created in debug_mode"""
        invoice = cfdv32.get_invoice(
            self.dict_invoice_basic_32_errored, debug_mode=True)
        self.assertTrue(invoice.document,
                        'A invalid dictionary gave error debugged_mode '
                        'enabled gave an error.')

    def test_006_download_file(self):
        """With a file it is downloaded and cached in a temporary file"""
        # TODO: Mock this
        downloaded = tools.cache_it(
            'http://s3.vauxoo.com/cadenaoriginal_3_2/cadenaoriginal_3_2.xslt')
        content = open(downloaded).read()
        self.assertTrue(content.find('se establece que la salida') > 0,
                        'I read the content of a cached file and '
                        'the result was not correct.')

    def test_007_cache(self):
        """With a file it is downloaded and cached in a temporary file"""
        # TODO: Mock this
        downloaded = tools.cache_it(
            'http://s3.vauxoo.com/cadenaoriginal_3_2/cadenaoriginal_3_2.xslt')
        content_xslt = downloaded
        content_xml = self.real_document_xml
        converted = tools.get_original(content_xml, content_xslt)
        self.assertTrue(
            converted,
            'I read the content of a cached file and the result '
            'was not correct.')
        self.assertTrue(len(tools.cached) > 1,
                        'Cache dictionary was not cached properly')


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_cfdilib
----------------------------------

Tests for `cfdicli` module.
"""

import unittest
from os.path import dirname, join
from shutil import copy

from click.testing import CliRunner

import cfdilib
from cfdilib.cfdicli import cli


class TestCFDIcli(unittest.TestCase):

    def _get_test_data_path(self, file_name):
        return join(self.test_data_dir, file_name)

    def setUp(self):
        self.test_data_dir = join(
            dirname(cfdilib.__file__), "..", "tests", "demo")

    def tearDown(self):
        pass

    def test_001_cli(self):
        """The cli interface works as expected,
        with correct and errored files"""

        runner = CliRunner()
        with runner.isolated_filesystem():
            test_path_ok = self._get_test_data_path('basic_invoice_32.txt')
            copy(test_path_ok, 'document.json')
            test_path_nook = self._get_test_data_path(
                'basic_invoice_32_errored.txt')
            result_ok = runner.invoke(
                cli, ['--in_file', test_path_ok, '--out_file',
                      '/tmp/document_ok.xml', 'cfdv32mx'])
            self.assertTrue(
                result_ok.output.find('has been created') > 0,
                'Output from cli different from expected when test file Ok.')
            result_nook = runner.invoke(
                cli, ['--in_file', test_path_nook, '--out_file',
                      '/tmp/document_nook.xml', 'cfdv32mx'])
            self.assertTrue(result_nook.output.find('Emisor') > 0,
                            'Output from cli different from expected when test'
                            ' file errored.')
            result_ok_nofiles = runner.invoke(cli, ['cfdv32mx'])
            self.assertTrue(
                result_ok_nofiles.output.find('has been created') > 0,
                'Output with no files as parameters failed.')



# coding: utf-8
'''Command line interface for cfdilib
'''

import click

from . import cfdv32


class Config(object):

    def __init__(self):
        self.verbose = False

pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group()
@click.option('--verbose', is_flag=True,
              help='Pass the result to the standard out')
@click.option('--out_file', type=click.File('wb'),
              default='document.xml')
@click.option('--in_file', type=click.File('rb'),
              default="document.json")
@pass_config
def cli(config, in_file, out_file, verbose):
    """Main Interface to generate xml documents
    from custom dictionaries using legal xsd files
    complying with legal documents in all countires
    around the world.
    """
    config.out_file = out_file
    config.verbose = verbose
    config.in_file = in_file
    config.out_file = out_file


@cli.command()
@pass_config
def cfdv32mx(config):
    """Format cfdi v3.2 for Mexico.

    \b
    File where the files will be written document.xml.
        cfdicli --in_file /path/to/yout/json/documnt.json cfdv32mx

    \b
    File where the files will be written from document.json.
        cfdicli --out_file ./document.xml cfdv32mx
    """
    # TODO: look for a secure option for eval.
    #       Or simply the CLI only should manage json?
    # TODO: Implement json option also.
    dict_input = eval(config.in_file.read())
    invoice = cfdv32.get_invoice(dict_input)
    if invoice.valid:
        config.out_file.write(invoice.document)
        config.out_file.flush()
        click.echo('Document %s has been created.' % config.out_file.name)
    else:
        click.echo(invoice.ups.message)



# -*- coding: utf-8 -*-

from .cfdilib import BaseDocument


class CFDv32(BaseDocument):
    """Invoice document to comply with
    cfdi: v3.2 for Invoice Mexico Standards."""

    def __init__(self, dict_invoice, debug_mode=False):
        self.template_fname = 'cfdv32.xml'
        self.xslt_fname = \
            'http://www.sat.gob.mx/sitio_internet/cfd/3/cadenaoriginal_3_2/cadenaoriginal_3_2.xslt'  # noqa
        self.global_namespace = 'http://www.sat.gob.mx/sitio_internet/cfd'
        self.set_template(self.template_fname)
        super(CFDv32, self).__init__(dict_invoice, debug_mode=debug_mode)
        # This method must be called from the inherited
        #   __init__ always AFTER the super.

    def set_template(self, template_fname):
        self.template = super(CFDv32, self).set_template(template_fname)

    def set_schema(self, schema_fname):
        self.schema = super(CFDv32, self).set_schema(schema_fname)

    def set_xslt(self):
        # TODO: Standarize the schema in this way also,
        #       we can not use different algorithms here
        self.xstl = super(CFDv32, self).set_xslt()


def get_invoice(dict_invoice, debug_mode=False):
    return CFDv32(dict_invoice, debug_mode=debug_mode)



# -*- coding: utf-8 -*-

from urllib2 import URLError

from suds.client import Client


class Signatory(object):
    """A third party element against which we will sign the invoice.
    In Mexico for example it is called PAC, in Spain is the goverment itself
    In other countries it can be a different entity, private or public and use
    different kind of webservices technologies.

    This is an object to be used as a helper, this should be overwritten
    if the way we decided to sign (SOAP, xmlrpc, json, etc, etc)
    change in your country.

    The default behavior is for the way Mexico do things.
    """
    def __init__(self, url, user, password):
        self.message = None
        self.client = None
        self.url = url
        self.user = user
        self.password = password
        self.document = None
        self.result = None

    def _sign(self):
        try:
            # The name of the service here will
            # change depending of the signatory third party.
            self.result = self.client.service.stamp(
                self.document, self.user, self.password)
        except Exception as e:
            print e  # pylint: disable=print-statement

    def sign(self, xml_doc):
        """Sign the document with a third party signatory.

        :param str xml_doc: Document self signed in plain xml
        :returns answer: Answer is given from the signatory
            itself if connected.
        """
        try:
            self.client = Client(self.url)
        except ValueError as e:
            self.message = e.message
        except URLError:
            self.message = 'The url you provided: ' + \
                '%s could not be reached' % self.url
